//X-10 Demo V1y
//RA (11110011)
//0   ZERO-Crossing
//1   X10-RX
//2   X10-OUTPUT
//3   UHF-TX
//4   UHF-RX
//5   (Any input)
//
//RB (11000010)
//0   X10-TX
//1   RS232 RX
//2   RS232 TX
//3   X10EX   (X10-Lamp/IR)
//4   Speaker
//5   LED-OK
//6   SW-ON
//7   SW-OFF

#include <16F648A.h>
//#include <16F628.h>
#include <string.h>
#include <stdlib.h>

#use delay(clock=4000000)
#use rs232(baud=9600,parity=N,xmit=PIN_B2,rcv=PIN_B1,bits=8,FORCE_SW)
#fuses XT,PUT,BROWNOUT,MCLR,NOWDT,NOPROTECT,NOLVP

#case
#use fast_io(A)
#define X10_ZERO_CROSS  PIN_A0
#define X10READPIN      PIN_A1
#define X10OUT          PIN_A2
#define X10UHFTX        PIN_A3
#define X10UHFRX        PIN_A4

#define X10WRITEPIN     PIN_B0
#define X10EX           PIN_B3
#define SPEAKER         PIN_B4
#define LEDOK           PIN_B5
#define SWOFF           PIN_B6
#define SWON            PIN_B7

#define X10WAIT         800
#define X10_STARTBIT    0B11100000

//X-10 Function Code
#define X10FN_ALL_UN_OFF    1
#define X10FN_ALL_UN_ON     2
#define X10FN_ON            3
#define X10FN_OFF           4
#define X10FN_DIM           5
#define X10FN_BRIGHT        6
#define X10FN_ALL_LG_OFF    7
#define X10FN_EXT_CODE      8
#define X10FN_HAIL_REQUEST  9
#define X10FN_HAIL_ACK      10
#define X10FN_PRE_SET_DIM   11
#define X10FN_EXT_CODE_AG   12
#define X10FN_STATUS_ON     13
#define X10FN_STATUS_OFF    14
#define X10FN_STATUS_REQ    15

#define PORTA               0x05
#define PORTB               0x06

#define KEYDELAY            50
#define KEY_ON              0x80
#define KEY_OFF             0x40
#define CMDSIZE             6

//Sound
#define   Speaker           PORTB         //Speaker Port
#define   Spkbit            0B00010000      //Speaker bit is   ra.0

#rom  0x2100={'A',0x04}        //Init EEPROM

//Compliment of House & Keycode (is same)
int const X10NumCode[16] ={0x69,0xA9,0x59,0x99,0x56,0x96,0x66,0xA6,0x6A,0xAA,0x5A,0x9A,0x55,0x95,0x65,0xA5};
//House Code
char const X10HouseCode[16]={'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P'};
//Compliment of Function Code 
int const X10FnCode[15] ={0x55,0x56,0x59,0x5A,0x65,0x66,0x69,0x6A,0x95,0x96,0x99,0xA5,0xA6,0xA9,0xAA};
//char const StrFn[12][22]={"ALL Units Off","ALL Units On","On","Off","Dim","Bright","All Lights Off","Extended Code","Hail Request","Hail Acknowledge",
//"Pre-set Dim","Extended Code(Analog)","Status = On","Status = Off","Status Request"};

int StartCode,HouseCode[2],KeyCode[2],kType[2];
int rHouse,rNumber,rFunction;
char StrFunc[22];
char PKStart,RxHouse;
char CmdBuff[CMDSIZE+1];
int PKIndex;
int RxNo,RxFn;

//&sound
//*** Sound ***
//Working send out put to toggle Spkbit
//XTAL 4MHz
void Sound(char freq,char dura)
{
int dl,cl;
 #asm

   MOVLW   50
   MOVWF   dl
SndNx:
   MOVF   freq,W
   MOVWF   cl
   MOVLW   Spkbit
   XORWF   Speaker,F
SndLp:
   NOP
   NOP
   NOP
   DECFSZ   cl,F
   GOTO   SndLp

   DECFSZ   dl,F
   GOTO   SndNx

   MOVLW   50
   MOVWF   dl
   DECFSZ   dura,F
   GOTO   SndNx

    //BCF     Speaker,7       //Clear Speaker bit low
 #endasm
// output_low(PIN_B7);
}
void SoundOK(void)
{
 Sound(25,20);
}
void SoundHL(void)
{
 Sound(25,75);
}
void SoundFn(void)
{
 Sound(20,35);
}
void SoundERR(void)
{
 Sound(210,10);
}
void SoundSuccess(void)
{
 Sound(25,10);
 delay_ms(50);
 Sound(25,10);
 delay_ms(50);
 Sound(25,10);
}
void SoundPress(void)
{
 Sound(35,10);
}

//END Sound

char ReadPortA(void)
{
 #asm
    movf  PORTA,0     //mov to W
    movwf _RETURN_
 #endasm
}
char ReadPortB(void)
{
 #asm
    movf  PORTB,0     //mov to W
    movwf _RETURN_
 #endasm
}
//X-10 Function *************************************
//Store Function detail in StrFunc[];
//Input number of Function 1-n
void StrFuncMsg(int n)
{
      switch(n)
      {
     case 1:
            strcpy(StrFunc,"ALL Units Off");
          break;
     case 2:
            strcpy(StrFunc,"ALL Units On");
          break;
     case 3:
            strcpy(StrFunc,"ON");
          break;
     case 4:
            strcpy(StrFunc,"OFF");
          break;
     case 5:
            strcpy(StrFunc,"Dim");
          break;
     case 6:
            strcpy(StrFunc,"Bright");
          break;
     case 7:
            strcpy(StrFunc,"All Lights Off");
          break;
     case 8:
            strcpy(StrFunc,"All Lights Off");
          break;
     case 9:
            strcpy(StrFunc,"Hail Request");
          break;
     case 10:
            strcpy(StrFunc,"Hail Acknowledge");
          break;
     case 11:
            strcpy(StrFunc,"Pre-set Dim");
          break;
     case 12:
            strcpy(StrFunc,"Extended Code(Analog)");
          break;
     case 13:
            strcpy(StrFunc,"Status = On");
          break;
     case 14:
            strcpy(StrFunc,"Status = Off");
          break;
     case 15:
            strcpy(StrFunc,"Status Request");
          break;
     default:
          strcpy(StrFunc,"Unknow");
      }
}
//--------------------------------------
void InitialChip(void)
{
   setup_comparator(NC_NC_NC_NC);   //Input Digital
   output_a(0);
   output_b(0B00000100);
   set_tris_a(0B11110011);
   set_tris_b(0B11000010);  
}

//Find House
char FindHouseCode(int H)
{
int i;

    for(i=0;i<16;i++)
    {
     if(H==X10NumCode[i])
       return X10HouseCode[i];
    }
  return 'X';
}
int FindKeyCode(int K)
{
int i;

    for(i=0;i<16;i++)
    {
     if(K==X10NumCode[i])
       return (i+1);
    }
  return 0;
}
//Find Function Code 
//Return    0     not found
//          (i+1) found in Table
int FindFnCode(int F)
{
int i;

    for(i=0;i<15;i++)
    {
     if(F==X10FnCode[i])
     {
       //printf("-%X-",X10FnCode[i]);
       return (i+1);
     }
    }
  return 0;
}
//Check bit 1 of Start code  1110
//Return 0 is not found, 1 is found
int X10READY(void)
{
 return(input(X10READPIN));
}
//Check ZeroCrossing 
void CheckZeroCross() 
{
  if(input(X10_ZERO_CROSS))             //If = 1
     while(input(X10_ZERO_CROSS)) ;     //wait 0
  else
     while(!input(X10_ZERO_CROSS)) ;    //if = 1 wait 0
}

//Read Data n bit form AC-Line Sync with ZeroCrossing
//Input  n
//Output XData
int X10ReadBit(int n) 
{
int xdata,i;

   xdata=0;
   for(i=0;i<n;++i) 
   {
    xdata |= input(X10READPIN);     //Save bit
    CheckZeroCross();
    delay_us(X10WAIT);
    if(i!=(n-1))                    //Check done
     xdata <<=1;                    //Shift Left one
   }
  return xdata;
}
//Check Start Code if good
//Read n Record in to buffer HouseCode[],KeyCode[],kType;
int X10ReadBuff(int n) 
{
int i;

  //Start Code
    for(i=0;i<n;i++)
    {
     StartCode=X10ReadBit(4);
     if(StartCode!=0B00001110)
      return 0;

     HouseCode[i]=X10ReadBit(8);
     KeyCode[i]=X10ReadBit(8);
     kType[i]=X10ReadBit(2);
    }

  return 1;
}

//Read Record
//Return 0  not Valid
//       1  Valid Record and Get Data in rHouse,rNumber,rFunction and get StrFunc
int X10Read(void)
{
int i,R;

    R=0;
    if(X10ReadBuff(2))
    {
     for(i=0;i<2;i++)
     {
      switch(kType[i])
      {                              
      //Address (Get HouseCode and NumberCode)
      case 1:
           output_high(LEDOK);
           if(i==1)
           {
            if(HouseCode[0]==HouseCode[1] && KeyCode[0]==KeyCode[1] && kType[0]==kType[1])
            {
             rHouse=FindHouseCode(HouseCode[1]);
             rNumber=FindKeyCode(KeyCode[1]);
            }
           }
          break;
      //Function (Get HouseCode and FunctionCode)
      case 2:    
          if(i==1)
          {
           if(HouseCode[0]==HouseCode[1] && KeyCode[0]==KeyCode[1] && kType[0]==kType[1])
           {
            rFunction=FindFnCode(KeyCode[1]);
            StrFuncMsg(rFunction);
            output_low(LEDOK);
            R=1;
           }
          }
          break;
      }
     }

    }

  return R;
}
//X10 write 1 bit 3 time Sync with Zero Crossing
void X10WriteBit(int b) 
{
int i;
//@
   CheckZeroCross();
   delay_us(500);
   
   for(i=0;i<3;i++) 
   {
    output_bit(X10WRITEPIN,b);
    delay_us(1400);
    output_low(X10WRITEPIN);
    if(i<2)
      delay_ms(2);
   }
}
//X10 write 1 byte (MSB write)
//Input Data  data
//      nBit
void X10WriteByte(int Data,int nBit) 
{
int i;
BOOLEAN xBit;

   for(i=0;i<nBit;++i) 
   {
    xBit=shift_left(&Data,1,0);
    X10WriteBit(xBit);
   }
}
//X10 write Block
//Input is  House,Number,Function
void X10Write(char House,int Number,int Fn)
{
int HouseCode,i;
    
    //printf("<\r\n");
    HouseCode=House-0x41;                       //Change to index
  //Address+Number twice
    for(i=0;i<2;i++)
    {
     X10WriteByte(X10_STARTBIT,4);              //Start bit
     X10WriteByte(X10NumCode[HouseCode],8);     //House A-P
     X10WriteByte(X10NumCode[Number-1],8);      //Number
     X10WriteByte(0B01000000,2);                //Alway 01
    }
    X10WriteByte(0B00000000,6);                 //Gap
  //Address+Function twice
    for(i=0;i<2;i++)
    {
     X10WriteByte(X10_STARTBIT,4);              //Start bit
     X10WriteByte(X10NumCode[HouseCode],8);     //House 
     X10WriteByte(X10FnCode[Fn-1],8);           //Function
     X10WriteByte(0B10000000,2);                //Alway 10
    }
}
//X10 write Only
//Input is  House,Function
void X10WriteFn(char House,int Fn)
{
int HouseCode,i;
    
    //printf("<\r\n");
    HouseCode=House-0x41;                       //Change to index
  //Address+Function twice
    for(i=0;i<2;i++)
    {
     X10WriteByte(X10_STARTBIT,4);              //Start bit
     X10WriteByte(X10NumCode[HouseCode],8);     //House 
     X10WriteByte(X10FnCode[Fn-1],8);           //Function
     X10WriteByte(0B10000000,2);                //Alway 10
    }
}
//Execute Function code
void X10FuncExe(int House,int Number,int Fn)
{
int H,N,P;
    
   //Get Address form EEPROM House & Number
    H=read_eeprom(0);
    N=read_eeprom(1);

//Private Function
    if(H==House && N==Number)
    {
     switch(Fn)
     {
      case X10FN_ON:                        //ON
        output_high(X10OUT);
        break;
      case X10FN_OFF:                       //OFF
        output_low(X10OUT);  
        break;
      case X10FN_STATUS_REQ:                     //12 ON ,13 OFF
          P=ReadPortA();
          P&=0B00000100;
          if(P==0B00000100)
             X10Write(H,N,X10FN_STATUS_ON);      //X10Write(H,N,12);
          else
             X10Write(H,N,X10FN_STATUS_OFF);
          break;
     }
    }

//ALL Unit Function
   switch(Fn)
   {
    case X10FN_ALL_UN_OFF:                  //All Units Off
    case X10FN_ALL_LG_OFF:                  //All Lights Off
      output_low(X10OUT);  
      break;

    case X10FN_ALL_UN_ON:                   //All Units On
      output_high(X10OUT);
        break;
   }

}
//Util***********************************

int ScanKey(void)
{
int D1,D2,Key;
int Cn;

   Key=0;
   D1=ReadPortB();              //Read again
   D1&=0B11000000;
   
   if(D1==0)                    //No Key press
    return 0;
   else
   {
    delay_ms(KEYDELAY);         //Delay for test
    D2=ReadPortB();             //Read again
    D2&=0B11000000;
    if(D1==D2)
    {
     Key=1;                     //Key press is normal
     Cn=0;
     do
     {
      //restart_wdt();
      delay_ms(30);
      D2=ReadPortB();           //Read again
      D2&=0B11000000;
      if(Key==1)                //Normal press only
      {
       Cn++;
       if(Cn==100)
       {
        output_low(X10OUT); 
        SoundFn();
        Key=2;                   //Key press is FnKey 
        Cn=0;
       }
      }
      //while(D2!=0 && Key==2);   //wait for FnKey Release
     }
     while(D2!=0);              //Wait for Release Key

     if(Key==1)
     {
       switch(D1)
       {
        case KEY_ON:
            output_high(X10OUT);
            break;
        case KEY_OFF:
            output_low(X10OUT);  
            break;
       }
      SoundPress();
     }
     
    }
   }

 return Key;
}
void SetAddress(void)
{
unsigned int32 Cn;
    
    Cn=0;
    while(TRUE)
    {
     if(X10READY())
     {
      if(X10Read())
      {
       write_eeprom (0,rHouse);
       write_eeprom (1,rNumber);
       printf("NEW ADDRESS=%C%02d\r\n",read_eeprom(0),read_eeprom(1));
       SoundSuccess();
       break;
      }
     }

     Cn++;                  //wait for 5 sec
     if(Cn>=250000)
     {
      SoundERR();
      printf("Setup Abort\r\n");
      break;
     }
    }

}
void PrintHelp(void)
{
    printf("\r\nEnter House,Number-Function(6 Character)\r\n");
    printf("Format HNN-FF\r\n");
    printf("   H=House in upper case (A-P)\r\n");   
    printf("   NN=Number with Zero (01-16)\r\n");
    printf("   FF is Function with Zero (01-15)\r\n");

    printf("\r\nExample-1\r\nA01-03<ENTER>  A01(ON)\r\n");
    printf("A01-04<ENTER>  A01(OFF)\r\n");

    printf("\r\nEnter House,Function(3 Character)\r\n");
    printf("Format HFF\r\n");
    printf("   H=House in upper case (A-P)\r\n");   
    printf("   FF is Function with Zero (01-15)\r\n");

    printf("\r\nExample-2\r\nA01<ENTER>  A(All Units Off)\r\n");
    printf("A05<ENTER>  A(Dim)\r\nA06<ENTER>  A(Bright)\r\n");
    printf("Number Function Valid is 1,2,5,6,7\r\n");
    printf("Pres ? for Help\r\n");

}
int CheckCmd3(char *Str)
{
int len,i,R;
char Buff[3];
    
    RxHouse=CmdBuff[0];
//Check Limit of Function
    R=1;
    if(CmdBuff[1]<'0'|| CmdBuff[1]>'9')
       R=0;
    else
    {
     if(CmdBuff[2]<'0'|| CmdBuff[2]>'9')
       R=0;
    }

//Check Value of Number & Function
    if(R==1)
    {
     //Test Function valid
      strncpy(Buff,CmdBuff+1,2);
      Buff[2]=0;
      RxFn=atoi(Buff);
      switch(RxFn)
      {
       case 1:      //All Units Off
       case 2:      //All Units On
       case 5:      //Dim
       case 6:      //Bright
       case 7:      //All Lights Off
           break;

       default:
            R=0;
      }
    }

 return R;
}
int CheckCmd6(char *Str)
{
int len,i,R;
char Buff[3];
    
    RxHouse=CmdBuff[0];
//Check Limit of Number & Function
    R=1;
    for(i=1;i<6;i++)
    {
     switch(i)
     {
      case 1:
      case 2:
      case 4:
      case 5:
          if(CmdBuff[i]<'0'|| CmdBuff[i]>'9')
           R=0;
          break;
      case 3:
          if(CmdBuff[i]!='-')
            R=0;
          break;
     }
    }

//Check Value of Number & Function
    if(R==1)
    {
     //Test number valid
      strncpy(Buff,CmdBuff+1,2);
      Buff[2]=0;
      RxNo=atoi(Buff);
      if(RxNo<1 || RxNo>16)
      {
       R=0;
       //printf("\r\nNumber Error\r\n");
      }
      else
      {
       strncpy(Buff,CmdBuff+4,2);
       Buff[2]=0;
       RxFn=atoi(Buff);
       if(RxFn<1 || RxFn>15)
       {
        R=0;
        //printf("\r\nFunction Error\r\n");
       }
      }
    }

 return R;
}
void UARTReady(void)
{
char c,R;
int len;
//signed int R;
    
    if(kbhit())
    {
      c = getc();
      if(PKStart==0)
      {
       if(c>='A'&& c<='P')
       {
        PKStart=1;
        PKIndex=0;
        CmdBuff[PKIndex]=c;
        PKIndex++;
        
        //printf("Start\r\n");
        printf("%c",c);
       }
       else
       {
        if(c=='?')
         PrintHelp();
        else
          printf("House code Error or not use Upper Letter\r\n");
       }
      }
      else
      {
       if(c=='\r'||c=='\n')         //check <ENTER> End command
       {
        CmdBuff[PKIndex]=0;
        printf("\r\n");

        len=strlen(CmdBuff);
        if(len!=6 && len!=3)
          printf("Command Length Error\r\n");
        else
        {
          //Check Record is Valid
            if(len==3)
              R=CheckCmd3(CmdBuff);
            else
            {
             if(len==6)
              R=CheckCmd6(CmdBuff);
            }
            if(R==0)
              printf("Command Error\r\n");
             else
            {
             if(len==3)    //House,Function
             {
              printf("Sending...%c%02d\r\n",RxHouse,RxFn);
              X10WriteFn(RxHouse,RxFn);
              if(RxFn==1||RxFn==2)
                X10FuncExe(RxHouse,RxNo,RxFn);
             }
             if(len==6)    //House,Number,Function
             {
              printf("Sending...%c%02d-%02d\r\n",RxHouse,RxNo,RxFn);
              X10Write(RxHouse,RxNo,RxFn);
              if(RxFn==3||RxFn==4)
                X10FuncExe(RxHouse,RxNo,RxFn);
             }
            }
        }
        PKStart=0;
        PKIndex=0;
       }
       else
       {
        printf("%c",c);
        CmdBuff[PKIndex]=c;        //store in buffer
        PKIndex++;
     //Test buffer full
        if(PKIndex==CMDSIZE+1)
        {
         CmdBuff[PKIndex]=0;
          //DoCmd(CmdBuff);
         //else
          //printf("Command Error\r\n");
         printf("Buffer Full\r\n");
         PKStart=0;
         PKIndex=0;
        }
       }
      }
    }
}
//int X10flag;
void main(void) 
{
int Key;

   InitialChip();
   output_high(LEDOK);
   delay_ms(500);
   output_low(LEDOK);
   printf("X10-LAB DEMO-2 ");
//default rHouse & rNumber
   rHouse=read_eeprom(0);
   rNumber=read_eeprom(1);
   printf("ADDRESS=%C%02d\r\n",rHouse,rNumber);
   printf("Press ? for Help\r\n");
   SoundSuccess();

   PKStart=0;
   PKIndex=0;
   while (TRUE) 
   {
     UARTReady();
    //Check X10 Line Status
     if(X10READY())
     {
      if(X10Read())
      {
       switch(rFunction)
       {
        case 1:
        case 2:
            printf("(Fn%02d)-%s\r\n",rFunction,StrFunc);
            break;
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
        case 9:
        case 10:
        case 11:
        case 12:
        case 13:
        case 14:
        case 15:
            //printf("%X %X %X  ",HouseCode[1],KeyCode[1],kType[1]); 
            printf("%c%02d(Fn%02d)-%s\r\n",rHouse,rNumber,rFunction,StrFunc);
            break;
        
        default:
            printf("UnKnow Function\r\n");
       }
       X10FuncExe(rHouse,rNumber,rFunction);
       SoundOK();
      }
     }

     Key=ScanKey();
     if(Key==2)
     {
      output_high(LEDOK);
      printf("Setup Address\r\n");
      SetAddress();
      output_low(LEDOK);
     }

   }
}

